
template <typename compo, typename alloc = std::allocator<compo> >
struct sub_matrix_base {
    using compo_type = compo;
    using alloc_type = alloc;

    virtual compo get (vec_si const &) const = 0;
    virtual compo get (sign_int const) const = 0;
    virtual compo const & con_ref (vec_si const &) const = 0;
    virtual compo const & con_ref (sign_int const) const = 0;

    virtual form const & get_form() const = 0;

    virtual void show (ostream & o) const {
        bool constexpr braces (false);

        form const & f {get_form ()};
        for (sign_int n {0}; n < f.count (); ++n) {
            vec_si const v {f.ylppa (n)};
            o << "\n    " << v << " = ";
            if (braces) o << "{ ";
            o << get(n);
            if (braces) o << " }";
        }
    }
    virtual ~sub_matrix_base () = default;
};

template <typename compo, typename alloc>
ostream & operator<< (ostream & o, sub_matrix_base<compo,alloc> const & mb)
{ mb.show(o); return o; }

template <typename compo, typename alloc = std::allocator<compo> >
struct var_matrix_base : public virtual sub_matrix_base<compo,alloc> {
    // even though this is named "var"_matrix_base, the
    // functions really are supposed to be declared as const
    virtual compo & var_ref (vec_si const &) const = 0;
    virtual compo & var_ref (sign_int const) const = 0;
    virtual void set (vec_si const &, compo const) const = 0;
    virtual void set (sign_int const, compo const) const = 0;

    template <typename input_iter>
    // C++ says that member template function cannot be virtual
    void set_all (input_iter begin, input_iter end) const {
        sign_int const count {this->get_form().count()};
        input_iter step {begin};
        sign_int dist {0};

        if (check) {
            while (true) {
                if (step == end) {
                    if (dist == count) break;
                    throw fail {"var_matrix_base::set_all ... iterator span smaller than count"};
                } else if (dist == count)
                    throw fail {"var_matrix_base::set_all ... count smaller than iterator span"};
                this->set (dist++, *(step++));
            }
        } else {
            while (true) {
                if (dist == count) break;
                this->set (dist++, *(step++));
            }
        }
    }
};

template <typename compo, typename alloc = std::allocator<compo> >
struct con_matrix_base : public virtual sub_matrix_base<compo,alloc> { };

template <typename compo, typename alloc>
void copy_block_base (
    sub_matrix_base<compo,alloc> const & srce_mat,
    form const & srce_inner_frm,
    var_matrix_base<compo,alloc> const & dest_mat,
    form const & dest_inner_frm
) {
    if (check) {
        form const & srce_outer_frm {srce_mat.get_form()};
        form const & dest_outer_frm {dest_mat.get_form()};

        if (! srce_outer_frm.contains(srce_inner_frm))
            throw fail {"copy_block_base ... srce inner form not contained in outer"};
        if (! dest_outer_frm.contains(dest_inner_frm))
            throw fail {"copy_block_base ... dest inner form not contained in outer"};
        if (srce_inner_frm.count() != dest_inner_frm.count())
            throw fail {"copy_block_base ... inner forms not same count"};
    }

    sign_int const count {srce_inner_frm.count()};
    for (sign_int n {0}; n < count; ++n) {
        vec_si const srce_index {srce_inner_frm.ylppa (n)};
        vec_si const dest_index {dest_inner_frm.ylppa (n)};
        dest_mat.set (dest_index, srce_mat.get (srce_index));
    }
}


