
template <typename compo, typename alloc>
class sub_matrix;

template <typename compo, typename alloc>
class var_matrix;

template <typename compo, typename alloc>
class con_matrix;

//////////////

template <typename compo, typename alloc>
bool sharing (
    sub_matrix<compo,alloc> const &,
    sub_matrix<compo,alloc> const &);

template <typename compo2, typename alloc2>
sub_matrix<compo2,alloc2> sub_matrix_linear (form const &, double const);

template <typename compo2, typename alloc2>
sub_matrix<compo2,alloc2> sub_matrix_random (form const &, rand_double const &);

//////////////

template <typename compo, typename alloc = std::allocator<compo> >
class sub_matrix : public virtual sub_matrix_base<compo,alloc> {
    using bod_com_all = body<compo,alloc>;
    form f;
    bod_com_all mutable b;
public:
    sub_matrix (sub_matrix const &) = default;
    sub_matrix (sub_matrix &&) = default;
    sub_matrix & operator= (sub_matrix const &) = default;
    sub_matrix & operator= (sub_matrix &&) = default;
    ~sub_matrix () = default;

    sub_matrix () :
        f { }, b (f.count (), compo {} ) { }

    sub_matrix (form const & of, compo const c = compo {}) :
        f {of}, b (f.count (), c) { }

    sub_matrix (form const & of, vector<compo,alloc> const & v) :
        f {of}, b {v}
    {
        if (check) if (f.count () != b.count ())
            throw fail {"sub_matrix::ctor ... form and vector different sizes"};
    }

    template <typename input_iter>
    sub_matrix (form const & of, input_iter begin, input_iter end) :
        f {of}, b (begin, end)
    {
        if (check) if (f.count () != b.count ())
            throw fail {"sub_matrix::ctor ... form and iterator span different sizes"};

    }

    sub_matrix (form const & of, initializer_list<compo> ilc) :
        f {of}, b {ilc}
    {
        if (check) if (f.count () != b.count ())
            throw fail {"sub_matrix::ctor ... form and vector different sizes"};
    }

    explicit sub_matrix (sub_matrix_base<compo,alloc> const & smb) :
        f {smb.get_form()}, b (f.count ())
    { for (sign_int n {0}; n < f.count(); ++n) b.set (n, smb.get (n)); }

    void assign_form (form const & o) {
        if (check) if (o.count () != f.count ())
            throw fail {"sub_matrix::assign ... wrong size"};
        f = o;
    }

    std::vector <compo,alloc> const & get_vector () const
    { return b.get_vector(); };

    void swap (sub_matrix & o) { f.swap (o.f); b.swap (o.b); }
    sign_int count () const { return b.count (); }
    form const & get_form () const override { return f; }
    
    friend bool sharing<> (sub_matrix const &, sub_matrix const &);

private:
    friend class var_matrix<compo,alloc>;
    friend class con_matrix<compo,alloc>;

    compo get (vec_si const & v) const override
    { return b.get (f.apply (v)); }
    compo get (sign_int const s) const override
    { return b.get (s); }

    compo const & con_ref (vec_si const & v) const override
    { return b.con_ref (f.apply (v)); }
    compo const & con_ref (sign_int const s) const override
    { return b.con_ref (s); }

    virtual compo & var_ref (vec_si const & v) const
    { return b.var_ref (f.apply (v)); }
    virtual compo & var_ref (sign_int const s) const
    { return b.var_ref (s); }

    virtual void set (vec_si const & v, compo const c) const
    { b.set (f.apply (v), c); }
    virtual void set (sign_int const s, compo const c) const
    { b.set (s, c); }

    template <typename compo2, typename alloc2>
    friend typename std::enable_if<std::is_floating_point<compo2>::value,sub_matrix<compo2,alloc2> >::type
        sub_matrix_linear2 (form const &, compo2 const);

    template <typename compo2, typename alloc2>
    friend typename std::enable_if<std::is_floating_point<compo2>::value,sub_matrix<compo2,alloc2> >::type
        sub_matrix_random2 (form const &, rand_double const &);
};

template <typename compo, typename alloc>
bool sharing (
    sub_matrix<compo,alloc> const & x,
    sub_matrix<compo,alloc> const & y)
{ return sharing<compo,alloc> (x.b, y.b); }
    
//////////////

template <typename compo, typename alloc = std::allocator<compo> >
struct var_matrix :
    public var_matrix_base<compo,alloc>,
    public sub_matrix<compo,alloc>
{
    using smcca = sub_matrix<compo,alloc>;

    var_matrix () = default;
    var_matrix (form const & of, compo const c = compo {}) : smcca {of, c} { }
    var_matrix (form const & of, vector<compo,alloc> const & v) : smcca {of, v} { }
    template <typename input_iter>
        var_matrix (form const & of, input_iter begin, input_iter end) :
            smcca {of, begin, end} { }
    var_matrix (form const & of, initializer_list<compo> ilc) : smcca {of, ilc} { }
    var_matrix (sub_matrix_base<compo,alloc> const & smb) :  smcca {smb} {}
    var_matrix (sub_matrix<compo,alloc> const & o) : smcca {o} { }

    compo get (vec_si const & v) const override
    { return smcca::get (v); }
    compo get (sign_int const s) const override
    { return smcca::get (s); }

    compo const & con_ref (vec_si const & v) const override
    { return smcca::con_ref (v); }
    compo const & con_ref (sign_int const s) const override
    { return smcca::con_ref (s); }

    // even though this is named "var"_matrix, the
    // next four functions really are const

    compo & var_ref (vec_si const & v) const override
    { return smcca::var_ref (v); }
    compo & var_ref (sign_int const s) const override
    { return smcca::var_ref (s); }

    void set (vec_si const & v, compo const c) const override
    { smcca::set (v, c); }
    void set (sign_int const s, compo const c) const override
    { smcca::set (s, c); }

    form const & get_form () const override { return smcca::get_form(); }
};

///////////

template <typename compo, typename alloc = std::allocator<compo> >
struct con_matrix :
    public con_matrix_base<compo,alloc>,
    public sub_matrix <compo,alloc>
{
    using smcca = sub_matrix<compo,alloc>;

    con_matrix () = default;
    con_matrix (form const & of, compo const c = compo {}) : smcca {of, c} { }
    con_matrix (form const & of, vector<compo,alloc> const & v) : smcca {of, v} { }
    template <typename input_iter>
        con_matrix (form const & of, input_iter begin, input_iter end) :
            smcca {of, begin, end} { }
    con_matrix (form const & of, initializer_list<compo> ilc) : smcca {of, ilc} { }
    con_matrix (sub_matrix_base<compo,alloc> const & smb) : smcca {smb} { }
    con_matrix (sub_matrix<compo,alloc> const & o) : smcca {o} { }
    con_matrix (var_matrix<compo,alloc> const & o) : smcca {o} { }

    compo get (vec_si const & v) const override
    { return smcca::get (v); }
    compo get (sign_int const s) const override
    { return smcca::get (s); }

    compo const & con_ref (vec_si const & v) const override
    { return smcca::con_ref (v); }
    compo const & con_ref (sign_int const s) const override
    { return smcca::con_ref (s); }

    form const & get_form () const override { return smcca::get_form(); }
};






