
template <typename compo, typename alloc>
class var_lens_permu_body;

template <typename compo, typename alloc>
class con_lens_permu_body;

/////////////

template <typename compo, typename alloc = std::allocator<compo> >
class var_lens_permu_body : public var_matrix_base<compo,alloc> {
    var_matrix_base<compo,alloc> const * mb;
    permu obv_per;
    permu inv_per;
    form f;  // should this be const * ?
    sign_int loc;
    sign_int bot;
public:
    var_lens_permu_body (var_matrix_base<compo,alloc> const & par_mb,
            permu const & par_per, sign_int par_loc) :
        mb {& par_mb},
        obv_per {par_per},
        inv_per {obv_per.invert()},
        f {mb->get_form()},
        loc {par_loc},
        bot {f.get(loc).get_bot()}
    {
        if (check) {
            sign_int const x {f.get(loc).get_len()};
            sign_int const y {obv_per.size()};
            if (x != y) throw fail {"var_lens_permu_body::ctor ... permu wrong size"};
        }
    }

    compo get (vec_si const & v) const override {
        vec_si u {v};
        u.at(loc) = obv_per.get(v.at(loc) - bot) + bot;
        return mb->get (u);
    }

    compo get (sign_int const s) const override {
        vec_si u {f.ylppa (s)};
        u.at(loc) = obv_per.get(u.at(loc) - bot) + bot;
        return mb->get (u);
    }

    compo const & con_ref (vec_si const & v) const override {
        vec_si u {v};
        u.at(loc) = obv_per.get(v.at(loc) - bot) + bot;
        return mb->con_ref (u);
    }

    compo const & con_ref (sign_int const s) const override {
        vec_si u {f.ylppa (s)};
        u.at(loc) = obv_per.get(u.at(loc) - bot) + bot;
        return mb->con_ref (u);
    }

    compo & var_ref (vec_si const & v) const override {
        vec_si u {v};
        u.at(loc) = obv_per.get(v.at(loc) - bot) + bot;
        return mb->var_ref (u);
    }

    compo & var_ref (sign_int const s) const override {
        vec_si u {f.ylppa (s)};
        u.at(loc) = obv_per.get(u.at(loc) - bot) + bot;
        return mb->var_ref (u);
    }

    void set (vec_si const & v, compo const c) const override {
        vec_si u {v};
        u.at(loc) = obv_per.get(v.at(loc) - bot) + bot;
        mb->set (u, c);
    }

    void set (sign_int const s, compo const c) const override {
        vec_si u {f.ylppa (s)};
        u.at(loc) = obv_per.get(u.at(loc) - bot) + bot;
        mb->set (u, c);
    }

    form const & get_form () const override { return f; }

    friend class con_lens_permu_body<compo,alloc>;
};

/////////////

template <typename compo, typename alloc = std::allocator<compo> >
class con_lens_permu_body : public con_matrix_base<compo,alloc> {
    sub_matrix_base<compo,alloc> const * mb;
    permu obv_per;
    permu inv_per;
    form f;
    sign_int loc;
    sign_int bot;
public:
    con_lens_permu_body (
            sub_matrix_base<compo,alloc> const & par_mb,
            permu const & par_per,
            sign_int par_loc) :
        mb {& par_mb},
        obv_per {par_per},
        inv_per {obv_per.invert()},
        f {mb->get_form()},
        loc {par_loc},
        bot {f.get(loc).get_bot()}
    {
        if (check) {
            sign_int const x {f.get(loc).get_len()};
            sign_int const y {obv_per.size()};
            if (x != y) throw fail {"con_lens_permu_body::ctor ... permu wrong size"};
        }
    }

    con_lens_permu_body (var_lens_permu_body<compo,alloc> const & o) :
        mb {o.mb},
        obv_per {o.obv_per}, inv_per {o.inv_per},
        f {o.f},
        loc {o.loc}, bot {o.bot}
    { }

    compo get (vec_si const & v) const override {
        vec_si u {v};
        u.at(loc) = obv_per.get(v.at(loc) - bot) + bot;
        return mb->get (u);
    }

    compo get (sign_int s) const override {
        vec_si u {f.ylppa (s)};
        u.at(loc) = obv_per.get(u.at(loc) - bot) + bot;
        return mb->get (u);
    }

    compo const & con_ref (vec_si const & v) const override {
        vec_si u {v};
        u.at(loc) = obv_per.get(v.at(loc) - bot) + bot;
        return mb->con_ref (u);
    }

    compo const & con_ref (sign_int s) const override {
        vec_si u {f.ylppa (s)};
        u.at(loc) = obv_per.get(u.at(loc) - bot) + bot;
        return mb->con_ref (u);
    }

    form const & get_form () const override { return f; }
};
