
template <typename compo, typename alloc>
class var_lens_less_range;

template <typename compo, typename alloc>
class con_lens_less_range;

/////////////

template <typename compo, typename alloc = std::allocator<compo> >
class var_lens_less_range : public var_matrix_base<compo,alloc> {
    var_matrix_base<compo,alloc> const * mb;
    form f;
public:
    var_lens_less_range (var_matrix_base<compo,alloc> const & par_mb, form const & par_frm) :
        mb {& par_mb}, f {par_frm}
    {
        if (check) if (! mb->get_form().contains(f))
            throw fail {"var_lens_less_range::ctor ... form not contained"};
    }

    compo get (vec_si const & v) const override {
        if (check) if (! f.contains(v))
            throw fail {"var_lens_less_range::get ... subscript out of range"};
        return mb->get (v);
    }
    compo get (sign_int const s) const override
    { return mb->get (f.ylppa (s)); }
  
    compo const & con_ref (vec_si const & v) const override {
        if (check) if (! f.contains(v))
            throw fail {"var_lens_less_range::con_ref ... subscript out of range"};
        return mb->con_ref (v);
    }
    compo const & con_ref (sign_int const s) const override
    { return mb->con_ref (f.ylppa (s)); }

    compo & var_ref (vec_si const & v) const override {
        if (check) if (! f.contains(v))
            throw fail {"var_lens_less_range::var_ref ... subscript out of range"};
        return mb->var_ref (v);
    }
    compo & var_ref (sign_int const s) const override
    { return mb->var_ref (f.ylppa (s)); }

    void set (vec_si const & v, compo const c) const override {
        if (check) if (! f.contains(v))
            throw fail {"var_lens_less_range::set ... subscript out of range"};
        mb->set (v, c);
    }
    void set (sign_int const s, compo const c) const override
    { mb->set (s, c); }

    form const & get_form () const override { return f; }

    friend class con_lens_less_range<compo,alloc>;
};

/////////////

template <typename compo, typename alloc = std::allocator<compo> >
class con_lens_less_range : public con_matrix_base<compo,alloc> {
    sub_matrix_base<compo,alloc> const * mb;
    form f;
public:
    con_lens_less_range (sub_matrix_base<compo,alloc> const & par_mb, form const & par_frm) :
        mb {& par_mb},
        f {par_frm}
    {
        if (check) if (! mb->get_form().contains(f))
            throw fail {"con_lens_less_range::ctor ... form not contained"};
    }

    con_lens_less_range (var_lens_less_range<compo,alloc> const & o) :
        mb {o.mb},
        f {o.get_form()}
    { }

    compo get (vec_si const & v) const override {
        if (check) if (! f.contains(v))
            throw fail {"con_lens_less_range::get ... subscript out of range"};
        return mb->get (v);
    }
    compo get (sign_int s) const override { return mb->get (f.ylppa (s)); }

    compo const & con_ref (vec_si const & v) const override {
        if (check) if (! f.contains(v))
            throw fail {"con_lens_less_range::con_ref ... subscript out of range"};
        return mb->con_ref (v);
    }
    compo const & con_ref (sign_int s) const override { return mb->con_ref (f.ylppa (s)); }

    form const & get_form () const override { return f; }
};

