
struct sub_lens_less_dimen;

template <typename compo, typename alloc>
class var_lens_less_dimen;

template <typename compo, typename alloc>
class con_lens_less_dimen;

/////////////

struct sub_lens_less_dimen {
    vec_si marker; 

    sub_lens_less_dimen (vec_si const & par_hold) :
        marker {par_hold}
    { }

    vec_si const & marker_get () const {
        return marker;
    }

    sign_int marker_get (sign_int const pos) const {
        if (! check) { return marker[pos]; }
        sign_int const ial {marker.at(pos)};
        if (ial == null_int)
            throw fail {"sub_lens_less_dimen::marker_get ... pos not part of marker"};

        return ial;
    }

    void marker_set (vec_si const & v) {
        marker = v;
    }

    void marker_set (sign_int const pos, sign_int const val) {
        if (! check) { marker[pos] = val; return; }
        sign_int & ial {marker.at(pos)};
        if (ial == null_int)
            throw fail {"sub_lens_less_dimen::marker_set ... pos not part of marker"};

        if (val == null_int)
            throw fail {"sub_lens_less_dimen::marker_set ... val is null_int"};
        marker[pos] = val;
    }

    void marker_add (sign_int const pos, sign_int const val) {
        if (! check) { marker[pos] += val; return; }
        sign_int & ial {marker.at(pos)};
        if (ial == null_int)
            throw fail {"sub_lens_less_dimen::marker_add ... pos not part of marker"};

        is_add_legal (ial, val);
        ial += val;
    }

    void marker_sub (sign_int const pos, sign_int const val) {
        if (! check) { marker[pos] += val; return; }
        sign_int & ial {marker.at(pos)};
        if (ial == null_int)
            throw fail {"sub_lens_less_dimen::marker_sub ... pos not part of marker"};

        is_sub_legal (ial, val);
        ial -= val;
    }

    sign_int marker_bot (sign_int const pos, form const & mbf) const {
        if (marker.at(pos) == null_int)
            throw fail{"sub_lens_less_dimen::marker_bot ... pos not part of marker" };
        return mbf.get(pos).get_bot();
    }

    sign_int marker_top (sign_int const pos, form const & mbf) const {
        if (marker.at(pos) == null_int)
            throw fail {"sub_lens_less_dimen::marker_top ... pos not part of marker" };
        return mbf.get(pos).get_top();
    }

    bool can_deref (form const & f) const {
        for (sign_int n {0}; n < marker.size(); ++n) {
            sign_int const ian {marker.at(n)};
            if (ian != null_int) {
                if (! f.get(n).contains(ian)) return false;
            }
        }
        return true;
    }
};

/////////////

template <typename compo, typename alloc = std::allocator<compo> >
class var_lens_less_dimen :
    public var_matrix_base<compo,alloc>,
    public sub_lens_less_dimen
{
    using baldc = sub_lens_less_dimen;

    var_matrix_base<compo,alloc> const * mb;
    form mbf;
    form f; // should this be const * ?

    form gather () const {
        vector<blt> temp;
        for (sign_int n {0}; n < baldc::marker.size(); ++n)
            if (baldc::marker.at(n) == null_int)
                temp.push_back (mbf.get(n));
        return form {temp};
    }

    vec_si full_subscript (vec_si const & v) const {
        if (check) if (v.size() != f.dimen())
            throw fail {"var_lens_less_dimen::full_subscript ... vec_si wrong size"};
        sign_int inner {0};
        vec_si work {baldc::marker};
        for (sign_int outer {0}; outer < work.size(); ++outer)
            if (work.at(outer) == null_int)
                work.at(outer) = v.at(inner++);
        return work;
    }
public:
    var_lens_less_dimen (
        var_matrix_base<compo,alloc> const & par_mb, vec_si const & par_hold) :
            sub_lens_less_dimen {par_hold},
            mb {& par_mb},
            mbf {mb->get_form()},
            f {gather()}
    {
        if (check) if (baldc::marker.size() != mbf.dimen())
            throw fail {"var_lens_less_dim::ctor ... marker wrong size"};
    }

    compo get (vec_si const & v) const override
    { return mb->get (full_subscript(v)); }
    compo get (sign_int const n) const override
    { return get (f.ylppa(n)); }

    compo const & con_ref (vec_si const & v) const override
    { return mb->con_ref (full_subscript(v)); }
    compo const & con_ref (sign_int const n) const override
    { return con_ref (f.ylppa(n)); }

    compo & var_ref (vec_si const & v) const override
    { return mb->var_ref (full_subscript(v)); }
    compo & var_ref (sign_int const n) const override 
    { return var_ref (f.ylppa(n)); }

    void set (vec_si const & v, compo const c) const override
    { mb->set (full_subscript(v), c); }
    void set (sign_int const s, compo const c) const override
    { set (f.ylppa (s), c); }

    form const & get_form () const override { return f; }

    vec_si const & get_marker () const { return baldc::marker; }

    sign_int marker_bot (sign_int const pos) const
    { return baldc::marker_bot (pos, mbf); }
    sign_int marker_top (sign_int const pos) const
    { return baldc::marker_top (pos, mbf); }

    bool can_deref () const { return baldc::can_deref (mbf); }

    friend class con_lens_less_dimen<compo,alloc>;
};

/////////////////////

template <typename compo, typename alloc = std::allocator<compo> >
class con_lens_less_dimen :
    public con_matrix_base<compo,alloc>,
    public sub_lens_less_dimen
{
    using baldc = sub_lens_less_dimen;

    sub_matrix_base<compo,alloc> const * mb;
    form mbf;
    form f;

    form gather () const {
        vector<blt> temp;
        for (sign_int n {0}; n < baldc::marker.size(); ++n)
            if (baldc::marker.at(n) == null_int)
                temp.push_back (mbf.get(n));
        return form {temp};
    }

    vec_si full_subscript (vec_si const & v) const {
        if (check) if (v.size() != f.dimen())
            throw fail {"con_lens_less_dimen::full_subscript ... vec_si wrong size"};
        sign_int inner {0};
        vec_si work {baldc::marker};
        for (sign_int outer {0}; outer < work.size(); ++outer)
            if (work.at(outer) == null_int)
                work.at(outer) = v.at(inner++);
        return work;
    }
public:
    con_lens_less_dimen (
        sub_matrix_base<compo,alloc> const & par_mb, vec_si const & par_hold) :
            baldc {par_hold},
            mb {& par_mb},
            mbf {mb->get_form()},
            f {gather()}
    {
        if (check) if (baldc::marker.size() != mbf.dimen())
            throw fail {"con_lens_less_dim::ctor ... marker wrong size"};
    }

    con_lens_less_dimen (var_lens_less_dimen<compo,alloc> const & o) :
        baldc {o.baldc::marker},
        mb {o.mb},
        mbf {o.mbf},
        f {o.f}
    { }

    compo get (vec_si const & v) const override
    { return mb->get (full_subscript(v)); }
    compo get (sign_int const n) const override
    { return get (f.ylppa(n)); }

    compo const & con_ref (vec_si const & v) const override
    { return mb->con_ref (full_subscript(v)); }
    compo const & con_ref (sign_int const n) const override
    { return con_ref (f.ylppa(n)); }

    form const & get_form () const override { return f; }

    vec_si const & get_marker () const { return baldc::marker; }

    sign_int marker_bot (sign_int const pos) const
    { return baldc::marker_bot (pos, mbf); }
    sign_int marker_top (sign_int const pos) const
    { return baldc::marker_top (pos, mbf); }

    bool can_deref () const { return baldc::can_deref (mbf); }
};


