
// scalar_non_var:

template <
    typename oper,
    typename compo_a,
    typename scalar_t = compo_a,
    typename alloc_a = std::allocator<compo_a>
> void scalar_non_var (
    var_matrix_base<compo_a,alloc_a> const & mat_a,
    scalar_t const scalar_v
) {
    form const & frm {mat_a.get_form()};
    sign_int const count {frm.count()};
    oper op;
    for (sign_int n {0}; n < count; ++n)
        op (mat_a.var_ref(n), scalar_v);
}

//////////////////////////////////////////////////

// scalar_ret_var:

template <
    typename oper,
    typename compo_a,
    typename scalar_t = compo_a,
    typename alloc_a = std::allocator<compo_a>
> var_matrix<compo_a,alloc_a> scalar_ret_var (
    var_matrix_base<compo_a,alloc_a> const & mat_a,
    scalar_t const scalar_v
) {
    form const & frm {mat_a.get_form()};
    sign_int const count {frm.count()};
    oper op;
    for (sign_int n {0}; n < count; ++n)
        op (mat_a.var_ref(n), scalar_v);
    return mat_a;
}

// operator*=:

template <
    typename compo_a, 
    typename scalar_t = compo_a,
    typename alloc_a = std::allocator<compo_a>
> var_matrix<compo_a,alloc_a> operator*= (
    var_matrix_base<compo_a,alloc_a> const & mat_a,
    scalar_t const scalar_v
) {
    struct mul_asgn {
        compo_a operator() (compo_a & a, scalar_t const b) { return a *= b; }
    };

    return scalar_ret_var <
        mul_asgn, compo_a, scalar_t, alloc_a
    > (mat_a, scalar_v);
}

// operator/=:

template <
    typename compo_a, 
    typename scalar_t = compo_a,
    typename alloc_a = std::allocator<compo_a>
> var_matrix<compo_a,alloc_a> operator/= (
    var_matrix_base<compo_a,alloc_a> const & mat_a,
    scalar_t const scalar_v
) {
    struct mul_asgn {
        compo_a operator() (compo_a & a, scalar_t const b) { return a /= b; }
    };

    return scalar_ret_var <
        mul_asgn, compo_a, scalar_t, alloc_a
    > (mat_a, scalar_v);
}

//////////////////////////////////////////////////

// scalar_ret_con:

template <
    typename oper,
    typename compo_a,
    typename scalar_t = compo_a,
    typename compo_z = compo_a,
    typename alloc_a = std::allocator<compo_a>,
    typename alloc_z = std::allocator<compo_z>
> var_matrix<compo_z,alloc_z> scalar_ret_con (
    sub_matrix_base<compo_a,alloc_a> const & mat_a,
    scalar_t const scalar_v
) {
    form const & frm {mat_a.get_form()};
    var_matrix<compo_z,alloc_z> mat_z {frm};
    sign_int const count {frm.count()};
    oper op;
    for (sign_int n {0}; n < count; ++n)
        mat_z.set (n, op (mat_a.get(n), scalar_v));
    return mat_z;
}

// binary operator*:

template <
    typename compo_a,
    typename scalar_t = compo_a,
    typename compo_z = compo_a,
    typename alloc_a = std::allocator<compo_a>,
    typename alloc_z = std::allocator<compo_z>
> var_matrix<compo_z,alloc_z> operator* (
    sub_matrix_base<compo_a,alloc_a> const & mat_a,
    scalar_t const scalar_v
) {
    struct mul {
        compo_z operator() (compo_a const a, scalar_t const b) { return a * b; }
    };

    return scalar_ret_con <
        mul, compo_a, scalar_t, compo_z, alloc_a, alloc_z
    > (mat_a, scalar_v);
}

// operator/:

template <
    typename compo_a,
    typename scalar_t = compo_a,
    typename compo_z = compo_a,
    typename alloc_a = std::allocator<compo_a>,
    typename alloc_z = std::allocator<compo_z>
> var_matrix<compo_z,alloc_z> operator/ (
    sub_matrix_base<compo_a,alloc_a> const & mat_a,
    scalar_t const scalar_v
) {
    struct mul {
        compo_z operator() (compo_a const a, scalar_t const b) { return a / b; }
    };

    return scalar_ret_con <
        mul, compo_a, scalar_t, compo_z, alloc_a, alloc_z
    > (mat_a, scalar_v);
}

