
// parallel_non_var_con:

template <
    typename oper,
    typename compo_a,
    typename compo_b = compo_a,
    typename alloc_a = std::allocator<compo_a>,
    typename alloc_b = std::allocator<compo_b>
> void parallel_non_var_con (
    var_matrix_base<compo_a,alloc_a> const & mat_a,
    sub_matrix_base<compo_b,alloc_b> const & mat_b
) {
    form const & frm {mat_a.get_form()};
    if (check) if (mat_b.get_form() != frm)
        throw fail {"parallel_non_var_con ... form mismatch"};
    sign_int const count {frm.count()};
    oper op;
    for (sign_int n {0}; n < count; ++n)
        op (mat_a.var_ref(n), mat_b.get(n));
}

//////////////////////////////////////////////////

// parallel_ret_var_con:

template <
    typename oper,
    typename compo_a,
    typename compo_b = compo_a,
    typename alloc_a = std::allocator<compo_a>,
    typename alloc_b = std::allocator<compo_b>
> var_matrix<compo_a,alloc_a> parallel_ret_var_con (
    var_matrix_base<compo_a,alloc_a> const & mat_a,
    sub_matrix_base<compo_b,alloc_b> const & mat_b
) {
    form const & frm {mat_a.get_form()};
    if (check) if (mat_b.get_form() != frm)
        throw fail {"parallel_ret_var_con ... form mismatch"};
    sign_int const count {frm.count()};
    oper op;
    for (sign_int n {0}; n < count; ++n)
        op (mat_a.var_ref(n), mat_b.get(n));
    return mat_a;
}

// operator+=:

template <
    typename compo_a,
    typename compo_b = compo_a,
    typename alloc_a = std::allocator<compo_a>,
    typename alloc_b = std::allocator<compo_b>
> var_matrix<compo_a,alloc_a> operator+= (
    var_matrix_base<compo_a,alloc_a> const & mat_a,
    sub_matrix_base<compo_b,alloc_b> const & mat_b
) {
    struct add_asgn {
        compo_a operator() (compo_a & a, compo_b const b) { return a += b; }
    };

    return parallel_ret_var_con <
        add_asgn, compo_a, compo_b, alloc_a, alloc_b
    > (mat_a, mat_b);
}

// operator-=:

template <
    typename compo_a,
    typename compo_b = compo_a,
    typename alloc_a = std::allocator<compo_a>,
    typename alloc_b = std::allocator<compo_b>
> var_matrix<compo_a,alloc_a> operator-= (
    var_matrix_base<compo_a,alloc_a> const & mat_a,
    sub_matrix_base<compo_b,alloc_b> const & mat_b
) {
    struct add_asgn {
        compo_a operator() (compo_a & a, compo_b const b) { return a -= b; }
    };

    return parallel_ret_var_con <
        add_asgn, compo_a, compo_b, alloc_a, alloc_b
    > (mat_a, mat_b);
}

//////////////////////////////////////////////////

// parallel_ret_con_con:

template <
    typename oper,
    typename compo_a,
    typename compo_b = compo_a,
    typename compo_z = compo_a,
    typename alloc_a = std::allocator<compo_a>,
    typename alloc_b = std::allocator<compo_b>,
    typename alloc_z = std::allocator<compo_z>
> var_matrix<compo_z,alloc_z> parallel_ret_con_con (
    sub_matrix_base<compo_a,alloc_a> const & mat_a,
    sub_matrix_base<compo_b,alloc_b> const & mat_b
) {
    form const & frm {mat_a.get_form()};
    if (check) if (mat_b.get_form() != frm)
        throw fail {"parallel_ret_con_con ... form mismatch"};
    var_matrix<compo_z,alloc_z> mat_z {frm};
    sign_int const count {frm.count()};
    oper op;
    for (sign_int n {0}; n < count; ++n)
        mat_z.set (n, op (mat_a.get(n), mat_b.get(n)));
    return mat_z;
}

// binary operator+:

template <
    typename compo_a,
    typename compo_b = compo_a,
    typename compo_z = compo_a,
    typename alloc_a = std::allocator<compo_a>,
    typename alloc_b = std::allocator<compo_b>,
    typename alloc_z = std::allocator<compo_z>
> var_matrix<compo_z,alloc_z> operator+ (
    sub_matrix_base<compo_a,alloc_a> const & mat_a,
    sub_matrix_base<compo_b,alloc_b> const & mat_b
) {
    struct add {
        compo_z operator() (compo_a const a, compo_b const b) { return a + b; }
    };

    return parallel_ret_con_con<
        add, compo_a, compo_b, compo_z, alloc_a, alloc_b, alloc_z
    > (mat_a, mat_b);
}

// binary operator-:

template <
    typename compo_a,
    typename compo_b = compo_a,
    typename compo_z = compo_a,
    typename alloc_a = std::allocator<compo_a>,
    typename alloc_b = std::allocator<compo_b>,
    typename alloc_z = std::allocator<compo_z>
> var_matrix<compo_z,alloc_z> operator- (
    sub_matrix_base<compo_a,alloc_a> const & mat_a,
    sub_matrix_base<compo_b,alloc_b> const & mat_b
) {
    struct add {
        compo_z operator() (compo_a const a, compo_b const b) { return a - b; }
    };

    return parallel_ret_con_con<
        add, compo_a, compo_b, compo_z, alloc_a, alloc_b, alloc_z
    > (mat_a, mat_b);
}

